import pandas as pd
from sqlalchemy import select, lambda_stmt, or_

from nuc_data_tool.db.base import Session
from nuc_data_tool.db.db_model import File, NucData, Nuc, PhysicalQuantity
from nuc_data_tool.utils.formatter import physical_quantity_list_generator, type_checker
from nuc_data_tool.utils.middle_steps import middle_steps_line_parsing


def fetch_files_by_name(filenames='all'):
    """
    根据输入的物理量名，从files table获取 file object(s)

    Parameters
    ----------
    filenames : str or list[str]

    Returns
    -------
    list[File]
    """
    stmt = lambda_stmt(lambda: select(File))
    if (filenames == 'all') or ('all' in list(filenames)):
        pass
    else:
        if isinstance(filenames, str):
            filenames = [filenames]

        stmt += lambda s: s.where(File.name.in_(filenames))

    try:
        with Session() as session:
            files = session.execute(stmt).scalars().all()
    finally:
        if not files:
            raise Exception(f"{filenames} doesn't exists")

    return files


def fetch_physical_quantities_by_name(physical_quantities):
    """
    根据输入的物理量名，从physical_quantities table获取 physical_quantity object(s)

    Parameters
    ----------
    physical_quantities : list[str] or str
        核素名，可以是核素名的list或str

    Returns
    -------
    list[PhysicalQuantity]

    See Also
    --------
    physical_quantity_list_generator : 根据输入生成对应的物理量list
    physical_quantity list
    """
    with Session() as session:
        physical_quantities_list = physical_quantity_list_generator(physical_quantities)
        stmt = (select(PhysicalQuantity)
                .where(PhysicalQuantity.name.in_(physical_quantities_list))
                )
        physical_quantities = session.execute(stmt).scalars().all()

    return physical_quantities


def fetch_data_by_filename_and_physical_quantities(filename, physical_quantities, is_all_step=False):
    """
    根据输入的File和physical quantities从Nuc， NucData，PhysicalQuantity table获取数据

    Parameters
    ----------
    filename : File
        File object
    physical_quantities : list[str] or str or list[PhysicalQuantity] or PhysicalQuantity
        物理量，可以是物理量名的list[str]或str，
        也可以是PhysicalQuantity list也可以是list[PhysicalQuantity]或PhysicalQuantity
    is_all_step : bool, default false
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------
    dict[str, pd.DataFrame]
        返回一个结果字典，key为物理量名(str)，value为对应物理量的数据(DataFrame)
    """
    dict_df_data = {}

    if type_checker(filename, File) == 'str':
        filename = fetch_files_by_name(filename).pop()

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    physical_quantity: PhysicalQuantity
    for physical_quantity in physical_quantities:
        nuc_data = fetch_data_by_filename_and_physical_quantity(filename,
                                                                physical_quantity,
                                                                is_all_step)
        dict_df_data[physical_quantity.name] = nuc_data

    return dict_df_data


def fetch_data_by_filename_and_physical_quantity(filename, physical_quantity, is_all_step=False):
    """
    根据输入的 File 和 physical quantity 从 Nuc， NucData，PhysicalQuantity table获取数据

    Parameters
    ----------
    filename : File
        File object
    physical_quantity : str or PhysicalQuantity
        物理量，可以是物理量的 str，PhysicalQuantity
    is_all_step : bool, default false
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------
    pd.DataFrame
    """

    if type_checker(filename, File) == 'str':
        filename = fetch_files_by_name(filename).pop()

    if type_checker(physical_quantity, PhysicalQuantity) == 'str':
        physical_quantity = fetch_physical_quantities_by_name(physical_quantity).pop()

    df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

    file_id = filename.id
    physical_quantity_id = physical_quantity.id

    if not is_all_step:
        # 不读取中间结果，所以不选择NucData.middle_steps，否则反之
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                          NucData.first_step, NucData.last_step))
    else:
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                          NucData.first_step, NucData.last_step,
                                          NucData.middle_steps))

    stmt += lambda s: s.join(Nuc,
                             Nuc.id == NucData.nuc_id)
    stmt += lambda s: s.join(PhysicalQuantity,
                             PhysicalQuantity.id == NucData.physical_quantity_id)
    stmt += lambda s: s.where(NucData.file_id == file_id,
                              PhysicalQuantity.id == physical_quantity_id)

    with Session() as session:
        column_names = [column.name for
                        column in list(stmt.selected_columns)]
        df_right = pd.DataFrame(data=session.execute(stmt).all(),
                                columns=column_names)
        if is_all_step:
            exclude_middle_steps = df_right.drop(columns='middle_steps', axis=1)
            del column_names[-1]
            exclude_middle_steps.columns = column_names

            middle_steps = pd.DataFrame([middle_steps_line_parsing(middle_steps)
                                         for middle_steps in df_right['middle_steps']
                                         if middle_steps is not None])

            df_right = pd.concat([exclude_middle_steps, middle_steps], axis=1, copy=False)

    if not df_right.empty:
        df_left = pd.merge(df_left, df_right, how='outer', on=['nuc_ix', 'name'])

    df_left.sort_values(by=['nuc_ix'], inplace=True)

    return df_left


def fetch_data_by_filename_and_nuclide_list(filename, physical_quantities, nuclide_list, is_all_step=False):
    """
    根据输入的File，physical quantities，nuclide_list(核素列表)，all_step
    从Nuc， NucData，PhysicalQuantity table获取数据

    Parameters
    ----------
    filename : File
        File object
    physical_quantities : list[str] or str or list[PhysicalQuantity] or PhysicalQuantity
        物理量，可以是物理量名的list[str]或str，
        也可以是PhysicalQuantity list也可以是list[PhysicalQuantity]或PhysicalQuantity
    nuclide_list : list[str]
        核素list
    is_all_step : bool, default false
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------
    dict[str, pd.DataFrame]
        返回一个结果字典，key为物理量名(str)，value为对应物理量的数据(DataFrame)
    """
    dict_df_data = {}

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    with Session() as session:
        physical_quantity: PhysicalQuantity
        for physical_quantity in physical_quantities:
            file_id = filename.id
            physical_quantity_id = physical_quantity.id

            if not is_all_step:
                # 不读取中间结果，所以不选择NucData.middle_steps，否则反之
                stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                                  NucData.first_step, NucData.last_step))
            else:
                stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                                  NucData.first_step, NucData.last_step,
                                                  NucData.middle_steps))

            stmt += lambda s: s.join(Nuc,
                                     Nuc.id == NucData.nuc_id)
            stmt += lambda s: s.join(PhysicalQuantity,
                                     PhysicalQuantity.id == NucData.physical_quantity_id)
            stmt += lambda s: s.where(NucData.file_id == file_id,
                                      PhysicalQuantity.id == physical_quantity_id)
            if nuclide_list is None:
                # 核素列表为空则过滤first_step和last_step皆为0的records
                stmt += lambda s: s.where(or_(NucData.first_step != 0, NucData.last_step != 0))
            else:
                if physical_quantity.name != 'gamma_spectra':
                    # 核素不为gamma时，依照核素列表过滤records，否则反之
                    stmt += lambda s: s.where(Nuc.name.in_(nuclide_list))

            nuc_data = pd.DataFrame(data=session.execute(stmt).all(),
                                    columns=tuple(column.name for
                                                  column in list(stmt.selected_columns))
                                    )

            if is_all_step:
                nuc_data_exclude_middle_steps = nuc_data.drop(columns='middle_steps', axis=1)
                middle_steps = pd.DataFrame([middle_steps_line_parsing(middle_steps)
                                             for middle_steps in nuc_data['middle_steps']
                                             if middle_steps is not None])

                del nuc_data
                nuc_data = pd.concat([nuc_data_exclude_middle_steps, middle_steps],
                                     axis=1, copy=False)

            nuc_data.sort_values(by=['nuc_ix'], inplace=True)
            dict_df_data[physical_quantity.name] = nuc_data

    return dict_df_data


def fetch_extracted_data_id(filenames=None, physical_quantities='all', nuclide_list=None):
    """
    获取extracted_data的id

    Parameters
    ----------
    filenames : list[File] or File
        File object
    physical_quantities : list[str or PhysicalQuantity] or str or PhysicalQuantity
        物理量，可以是物理量名的list[str]或str，
        也可以是list[PhysicalQuantity]或PhysicalQuantity
    nuclide_list : list[str]
        核素list

    Returns
    -------
    list[int]
    """

    if type_checker(filenames, File) == 'str':
        filenames = fetch_files_by_name(filenames)

    if not isinstance(filenames, list):
        filenames = [filenames]

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    nuc_data_id = []

    with Session() as session:
        for filename in filenames:
            physical_quantities_id = [physical_quantity.id
                                      for physical_quantity in physical_quantities]
            file_id = filename.id

            if nuclide_list is None:
                # 核素列表为空则过滤first_step和last_step皆为0的records
                stmt = (select(NucData.id).
                        where(NucData.file_id == file_id,
                              NucData.physical_quantity_id.in_(physical_quantities_id)).
                        where(or_(NucData.first_step != 0, NucData.last_step != 0))
                        )
            else:
                # 核素不为gamma时，依照核素列表过滤records，否则反之
                for physical_quantity in physical_quantities:
                    if physical_quantity.name == 'gamma_spectra':
                        gamma_physical_quantity_id = physical_quantity.id

                        gamma_stmt = (select(NucData.id).
                                      where(NucData.file_id == file_id,
                                            NucData.physical_quantity_id == gamma_physical_quantity_id)
                                      )
                        nuc_data_id.extend(session.execute(gamma_stmt).scalars().all())

                stmt = (select(NucData.id).
                        join(Nuc, Nuc.id == NucData.nuc_id).
                        where(NucData.file_id == file_id,
                              NucData.physical_quantity_id.in_(physical_quantities_id)).
                        where(Nuc.name.in_(nuclide_list))
                        )

            nuc_data_id.extend(session.execute(stmt).scalars().all())

    return nuc_data_id


def fetch_extracted_data_by_filename_and_physical_quantity(nuc_data_id,
                                                           filename,
                                                           physical_quantity,
                                                           is_all_step=False):
    """
    获取 extracted_data

    Parameters
    ----------
    nuc_data_id : list[int]
    filename :str or File
    physical_quantity : str or PhysicalQuantity
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------

    """

    if type_checker(filename, File) == 'str':
        filename = fetch_files_by_name(filename).pop()

    if type_checker(physical_quantity, PhysicalQuantity) == 'str':
        physical_quantity = fetch_physical_quantities_by_name(physical_quantity).pop()

    df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

    physical_quantity_id = physical_quantity.id

    filename: File
    file_id = filename.id

    if not is_all_step:
        # 不读取中间结果，所以不选择NucData.middle_steps，否则反之
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix,
                                          Nuc.name,
                                          NucData.last_step).
                           where(NucData.id.in_(nuc_data_id)))
    else:
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix,
                                          Nuc.name,
                                          NucData.last_step,
                                          NucData.middle_steps).
                           where(NucData.id.in_(nuc_data_id)))

    stmt += lambda s: s.join(Nuc,
                             Nuc.id == NucData.nuc_id)
    stmt += lambda s: s.join(PhysicalQuantity,
                             PhysicalQuantity.id == NucData.physical_quantity_id)
    stmt += lambda s: s.where(NucData.file_id == file_id,
                              PhysicalQuantity.id == physical_quantity_id)

    with Session() as session:
        if not is_all_step:
            column_names = ['nuc_ix', 'name', f'{filename.name}_last_step']
            df_right = pd.DataFrame(data=session.execute(stmt).all(),
                                    columns=column_names)
        else:
            column_names = ['nuc_ix', 'name', f'{filename.name}_last_step', 'middle_steps']
            df_right = pd.DataFrame(data=session.execute(stmt).all(),
                                    columns=column_names)

            exclude_middle_steps = df_right.drop(columns='middle_steps', axis=1)
            del column_names[-1]
            exclude_middle_steps.columns = column_names

            middle_steps = pd.DataFrame([middle_steps_line_parsing(middle_steps)
                                         for middle_steps in df_right['middle_steps']
                                         if middle_steps is not None])
            middle_step_column_names = [f'{filename.name}_{name}'
                                        for name in middle_steps.columns.tolist()]
            middle_steps.columns = middle_step_column_names

            df_right = pd.concat([exclude_middle_steps, middle_steps], axis=1, copy=False)

    if not df_right.empty:
        df_left = pd.merge(df_left, df_right, how='outer', on=['nuc_ix', 'name'])

    df_left.sort_values(by=['nuc_ix'], inplace=True)

    return df_left


def fetch_max_num_of_middle_steps(physical_quantity='isotope'):
    """
    获取选定物理量中所有文件 middle_step 的最大值

    Parameters
    ----------
    physical_quantity : str or PhysicalQuantity, default = 'isotope'
        物理量，可以是物理量名的list[str]或str，
        默认为核素密度

    Returns
    -------
    int
    """
    files = fetch_files_by_name(filenames='all')

    if type_checker(physical_quantity, PhysicalQuantity) == 'str':
        physical_quantity = fetch_physical_quantities_by_name(physical_quantity).pop()

    max_num = 0
    for file in files:
        nuc_data = fetch_data_by_filename_and_physical_quantity(file,
                                                                physical_quantity,
                                                                True)
        cur_num = len(nuc_data.columns)
        if max_num < cur_num:
            max_num = cur_num

    return max_num - 4


def fetch_transposed_data_by_filename_and_physical_quantity(filename,
                                                            physical_quantity,
                                                            is_all_step=False):
    """
    根据输入的 File 和 physical quantity 从 Nuc， NucData，PhysicalQuantity table获取数据

    Parameters
    ----------
    filename : File
        File object
    physical_quantity : str or PhysicalQuantity
        物理量，可以是物理量的 str，PhysicalQuantity
    is_all_step : bool, default false
        是否读取全部中间结果数据列，默认只读取第一步和最后一步

    Returns
    -------
    pd.DataFrame
    """

    if type_checker(filename, File) == 'str':
        filename = fetch_files_by_name(filename).pop()

    if type_checker(physical_quantity, PhysicalQuantity) == 'str':
        physical_quantity = fetch_physical_quantities_by_name(physical_quantity).pop()

    df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

    file_id = filename.id
    physical_quantity_id = physical_quantity.id

    if not is_all_step:
        # 不读取中间结果，所以不选择NucData.middle_steps，否则反之
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                          NucData.first_step, NucData.last_step))
    else:
        stmt = lambda_stmt(lambda: select(Nuc.nuc_ix, Nuc.name,
                                          NucData.first_step, NucData.last_step,
                                          NucData.middle_steps))

    stmt += lambda s: s.join(Nuc,
                             Nuc.id == NucData.nuc_id)
    stmt += lambda s: s.join(PhysicalQuantity,
                             PhysicalQuantity.id == NucData.physical_quantity_id)
    stmt += lambda s: s.where(NucData.file_id == file_id,
                              PhysicalQuantity.id == physical_quantity_id)

    with Session() as session:
        column_names = [column.name for
                        column in list(stmt.selected_columns)]
        df_right = pd.DataFrame(data=session.execute(stmt).all(),
                                columns=column_names)

    if is_all_step:
        exclude_middle_steps = df_right.drop(columns='middle_steps', axis=1)
        del column_names[-1]
        exclude_middle_steps.columns = column_names

        middle_steps = pd.DataFrame([middle_steps_line_parsing(middle_steps)
                                     for middle_steps in df_right['middle_steps']
                                     if middle_steps is not None])

        df_right = pd.concat([exclude_middle_steps, middle_steps], axis=1, copy=False)

    if not df_right.empty:
        df_left = pd.merge(df_left, df_right, how='outer', on=['nuc_ix', 'name'])

    df_left.sort_values(by=['nuc_ix'], inplace=True)

    nuc_ix = df_left.loc[:, ['nuc_ix', 'name']]

    df_left = df_left.T
    df_left.columns = df_left.loc['name']
    df_left.drop(['nuc_ix', 'name'], inplace=True)
    df_left = df_left.astype('float64', copy=False)
    reindex = df_left.index.tolist()
    reindex.append(reindex.pop(1))
    df_left = df_left.reindex(reindex, copy=False)

    time_interval = pd.Series((filename.time_interval * i for i in range(filename.repeat_times + 1)),
                              name='time_interval', index=df_left.index)
    time_interval = time_interval / pd.to_timedelta(1, unit='D')

    df_left = pd.concat([df_left, time_interval], axis=1)

    return nuc_ix, df_left



from pathlib import Path

import click

from nuc_data_tool import __version__
from nuc_data_tool.anomaly_detection.train_and_detection import save_prediction_to_exel
from nuc_data_tool.db.db_utils import init_db
from nuc_data_tool.db.fetch_data import (fetch_extracted_data_id,
                                         fetch_physical_quantities_by_name,
                                         fetch_files_by_name)
from nuc_data_tool.utils.configlib import config
from nuc_data_tool.utils.data_extraction import save_extracted_data_to_exel
from nuc_data_tool.utils.fill_db import populate_database
from nuc_data_tool.utils.formatter import (all_physical_quantity_list,
                                           physical_quantity_list_generator)
from nuc_data_tool.utils.input_xml_file import InputXmlFileReader
from nuc_data_tool.utils.relative_error_calculation import save_comparison_result_to_excel


class PythonLiteralOption(click.Option):

    def type_cast_value(self, ctx, value):
        return list(map(str.strip, value.strip('][').replace('"', '').split(',')))


class MutuallyExclusiveOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop('mutually_exclusive', []))
        help_args = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(self.mutually_exclusive)
            kwargs['help'] = help_args + (
                    ' NOTE: This argument is mutually exclusive with '
                    ' arguments: [' + ex_str + '].'
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(
                    self.name,
                    ', '.join(self.mutually_exclusive)
                )
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


@click.group()
@click.version_option(__version__)
def main_cli():
    """
    app 命令行
    """
    pass


@main_cli.command()
@click.option('--path', '-p',
              'path',
              default=config.get_file_path('test_file_path'),
              type=click.Path(),
              help='输出文件路径，默认读取配置文件中的路径')
@click.option('--physical_quantities', '-pq',
              'physical_quantities',
              default=all_physical_quantity_list,
              type=click.Choice(all_physical_quantity_list,
                                case_sensitive=False),
              multiple=True,
              help='物理量，默认为全部物理量')
@click.option('--initiation', '-init',
              'initiation',
              is_flag=True,
              default=False,
              help='初始化数据库')
def pop(path,
        physical_quantities,
        initiation):
    """
    将输出文件(*.xml.out) 的内容填充进数据库
    """

    if initiation is True:
        init_db()

    physical_quantities = physical_quantity_list_generator(physical_quantities)

    file_names = sorted(Path(path).glob('*.out'))
    for file_name in file_names:
        with InputXmlFileReader(file_name, physical_quantities) as xml_file:
            print(f'{xml_file.out_name}:')
            print(f'found:     {xml_file.fetched_physical_quantity}')
            print(f'not found: {xml_file.unfetched_physical_quantity}')
            print()
            populate_database(xml_file)


@main_cli.command()
@click.argument('filenames',
                nargs=-1)
@click.option('--result_path', '-p',
              'result_path',
              default=config.get_file_path('result_file_path'),
              type=click.Path(),
              help='输出文件路径，默认读取配置文件中的路径')
@click.option('--physical_quantities', '-pq',
              'physical_quantities',
              default=all_physical_quantity_list,
              type=click.Choice(all_physical_quantity_list,
                                case_sensitive=False),
              multiple=True,
              help='物理量，默认为全部物理量')
@click.option('--nuclide', '-n',
              'nuclide_list',
              default='fission_light',
              type=click.Choice(config.get_conf('nuclide_list').keys(),
                                case_sensitive=False),
              help='核素列表，从配置文件 nuclide_list 项下读取，默认 fission_light')
@click.option('--all_step', '-all',
              'is_all_step',
              is_flag=True,
              default=False,
              help='提取中间步骤')
@click.option('--merge', '-m',
              'merge',
              is_flag=True,
              default=False,
              help='将结果合并输出至一个文件')
def extract(filenames,
            result_path,
            physical_quantities,
            nuclide_list,
            is_all_step,
            merge):
    """
    从数据库导出选中的文件的数据到工作簿(xlsx文件)

    参数为文件列表(默认为所有文件)

    \b
    nuc_data_tool extract 'homo-case001-006' 'homo-case007-012' 'homo-case013-018'
    nuc_data_tool extract 'homo-case001-006'
    \b
    文件名(没有后缀) 例如：001.xml.out -> 001
    文件名列表 例如： 001 002 003
    """

    if filenames:
        filenames = fetch_files_by_name(filenames)
    else:
        filenames = fetch_files_by_name('all')

    physical_quantities = fetch_physical_quantities_by_name(physical_quantities)
    nuc_data_id = fetch_extracted_data_id(filenames,
                                          physical_quantities,
                                          config.get_nuclide_list(nuclide_list))

    save_extracted_data_to_exel(nuc_data_id=nuc_data_id,
                                filenames=filenames,
                                physical_quantities=physical_quantities,
                                is_all_step=is_all_step,
                                result_path=result_path,
                                merge=merge)


@main_cli.command()
@click.argument('reference_file',
                nargs=1)
@click.argument('comparison_files',
                nargs=-1)
@click.option('--result_path', '-p',
              'result_path',
              default=config.get_file_path('result_file_path'),
              type=click.Path(),
              help='输出文件路径，默认读取配置文件中的路径')
@click.option('--physical_quantities', '-pq',
              'physical_quantities',
              default=['isotope'],
              type=click.Choice(all_physical_quantity_list,
                                case_sensitive=False),
              multiple=True,
              help='物理量，默认为 isotope')
@click.option('--nuclide', '-n',
              'nuclide_list',
              default='fission_light',
              type=click.Choice(config.get_conf('nuclide_list').keys(),
                                case_sensitive=False),
              help='核素列表，从配置文件 nuclide_list 项下读取')
@click.option('--deviation_mode',
              '-dm',
              'deviation_mode',
              default='relative',
              type=click.Choice(['relative', 'absolute']),
              help='偏差模式，分为绝对和相对，默认为相对')
@click.option('--threshold',
              '-t',
              'threshold',
              default='1.0E-12',
              help='偏差阈值，默认1.0E-12')
@click.option('--all_step', '-all',
              'is_all_step',
              is_flag=True,
              default=False,
              help='提取中间步骤')
def compare(reference_file,
            comparison_files,
            result_path,
            physical_quantities,
            nuclide_list,
            deviation_mode,
            threshold,
            is_all_step):
    """
    \b
    对文件列表进行两两组合，进行对比，计算并输出对比结果至工作簿(xlsx文件)
    第一个参数为基准文件，第二个参数为文件列表(默认为除基准文件以外的所有文件)
    \b
    nuc_data_tool compare 'homo-case001-006' 'homo-case007-012' 'homo-case013-018'
    nuc_data_tool compare 'homo-case001-006'
    \b
    文件名(没有后缀) 例如：001.xml.out -> 001
    文件名列表 例如： 001 002 003
    """

    reference_file = fetch_files_by_name(reference_file).pop()

    if comparison_files:
        comparison_files = fetch_files_by_name(comparison_files)
    else:
        comparison_files = fetch_files_by_name('all')

    comparison_files = [comparison_file for comparison_file in comparison_files
                        if comparison_file.id != reference_file.id]

    physical_quantities = fetch_physical_quantities_by_name(physical_quantities)
    nuc_data_id = fetch_extracted_data_id([reference_file, *comparison_files],
                                          physical_quantities,
                                          config.get_nuclide_list(nuclide_list))

    save_comparison_result_to_excel(nuc_data_id=nuc_data_id,
                                    reference_file=reference_file,
                                    comparison_files=comparison_files,
                                    result_path=result_path,
                                    physical_quantities=physical_quantities,
                                    deviation_mode=deviation_mode,
                                    threshold=threshold,
                                    is_all_step=is_all_step)


@main_cli.command()
@click.argument('filenames',
                nargs=-1)
@click.option('--result_path', '-rp',
              'result_path',
              default=config.get_file_path('result_file_path'),
              type=click.Path(),
              help='输出文件路径，默认读取配置文件中的路径')
@click.option('--model_type', '-mt',
              'model_type',
              default='iforest',
              type=click.Choice(['abod', 'iforest', 'cluster',
                                 'cof', 'histogram', 'knn',
                                 'lof', 'svm', 'pca', 'mcd',
                                 'sod', 'sos'],
                                case_sensitive=True),
              help="""
‘abod’	Angle-base Outlier Detection
‘iforest’	Isolation Forest
‘cluster’	Clustering-Based Local Outlier
‘cof’	Connectivity-Based Outlier Factor
‘histogram’	Histogram-based Outlier Detection
‘knn’	k-Nearest Neighbors Detector
‘lof’	Local Outlier Factor
‘svm’	One-class SVM detector
‘pca’	Principal Component Analysis
‘mcd’	Minimum Covariance Determinant
‘sod’	Subspace Outlier Detection
‘sos	Stochastic Outlier Selection
""")
@click.option('--fraction',
              '-fra',
              'fraction',
              type=click.FLOAT,
              default=0.01,
              help='异常占数据集的比例，默认0.01')
@click.option('--model_path', '-mp',
              'model_path',
              default=config.get_anomaly_detection_config('model_path'),
              type=click.Path(),
              help='模型文件路径，默认读取配置文件中的路径')
@click.option('--physical_quantities', '-pq',
              'physical_quantities',
              default=['isotope'],
              type=click.Choice(all_physical_quantity_list,
                                case_sensitive=False),
              multiple=True,
              help='物理量，默认为 isotope')
@click.option('--all_step', '-all',
              'is_all_step',
              is_flag=True,
              default=False,
              help='提取中间步骤')
@click.option('--merge', '-m',
              'merge',
              is_flag=True,
              default=False,
              help='将结果合并输出至一个文件')
def detect(filenames,
           result_path,
           model_type,
           model_path,
           fraction,
           physical_quantities,
           is_all_step,
           merge):
    """
    对数据进行异常检测，并导出异常的数据至工作簿(xlsx文件)
    如果未输入model_type，model_path已输入，
    则会使用选择的预训练模型进行检测（请务必清楚自己正在干什么）

    参数为文件列表(默认为所有文件)
    """

    if filenames:
        filenames = fetch_files_by_name(filenames)
    else:
        filenames = fetch_files_by_name('all')

    physical_quantities = fetch_physical_quantities_by_name(physical_quantities)
    model_name = str(Path(model_path).parent.joinpath(Path(model_path).stem))

    save_prediction_to_exel(filenames=filenames,
                            result_path=result_path,
                            model_type=model_type,
                            fraction=fraction,
                            model_name=model_name,
                            physical_quantities=physical_quantities,
                            is_all_step=is_all_step,
                            merge=merge)


@main_cli.command()
@click.option('--file', '-f',
              'files',
              cls=MutuallyExclusiveOption,
              is_flag=True,
              default=False,
              mutually_exclusive=['physical_quantity'],
              help='显示文件信息')
@click.option('--physical_quantity', '-p',
              'physical_quantities',
              cls=MutuallyExclusiveOption,
              is_flag=True,
              default=False,
              mutually_exclusive=['files'],
              help='显示物理量信息')
@click.option('--list', '-l',
              'is_list',
              is_flag=True,
              default=False,
              help='以数组形式输出')
def fetch(files,
          physical_quantities,
          is_list):
    """
    获取 文件、物理量信息
    """

    if files is True:
        file_list = fetch_files_by_name('all')
        if is_list is False:
            for file in file_list:
                print(f'Name: {file.name}')
        else:
            print([file.name for file in file_list])

    if physical_quantities is True:
        physical_quantity_list = fetch_physical_quantities_by_name('all')
        if is_list is False:
            for physical_quantity in physical_quantity_list:
                print(f'Name: {physical_quantity.name}')
        else:
            print([physical_quantity.name for physical_quantity in physical_quantity_list])


def main():
    main_cli(prog_name='nuctool')


main()



from pathlib import Path

import pandas as pd
from openpyxl import load_workbook


def get_column_index(filename):
    """
    :param filename: input file
    :return: data rang of column ex: (2, 32)
    """
    start_column_index = 2
    wb = load_workbook(filename)
    sheet = wb.worksheets[0]
    # print(sheet)
    max_column = sheet.max_column
    # print(max_column)
    return start_column_index, max_column


def append_df_to_excel(filename, df, sheet_name='Sheet1', startrow=None,
                       truncate_sheet=False,
                       **to_excel_kwargs):
    """
    Append a DataFrame [df] to existing Excel file [filename]
    into [sheet_name] Sheet.
    If [filename] doesn't exist, then this function will create it.

    Parameters
    ----------
    filename : Path or str or pd.ExcelWriter
        File path or existing ExcelWriter
        (Example: '/path/to/file.xlsx')
    df : pd.DataFrame
        DataFrame to save to workbook
    sheet_name : str
        Name of sheet which will contain DataFrame.
        (default: 'Sheet1')
    startrow : int
        upper left cell row to dump data frame.
        Per default (startrow=None) calculate the last row
        in the existing DF and write to the next row...
    truncate_sheet : bool
        truncate (remove and recreate) [sheet_name]
        before writing DataFrame to Excel file
    to_excel_kwargs : kwarg
        arguments which will be passed to `DataFrame.to_excel()`
        [can be a dictionary]
    Returns
    -------

    """
    # Excel file doesn't exist - saving and exiting
    if not Path(filename).is_file():
        df.to_excel(
            filename,
            sheet_name=sheet_name,
            startrow=startrow if startrow is not None else 0,
            **to_excel_kwargs)
        return

    # ignore [engine] parameter if it was passed
    if 'engine' in to_excel_kwargs:
        to_excel_kwargs.pop('engine')

    writer = pd.ExcelWriter(filename, engine='openpyxl', mode='a')

    # try to open an existing workbook
    writer.book = load_workbook(filename)

    # get the last row in the existing Excel sheet
    # if it was not specified explicitly
    if startrow is None and sheet_name in writer.book.sheetnames:
        startrow = writer.book[sheet_name].max_row

    # truncate sheet
    if truncate_sheet and sheet_name in writer.book.sheetnames:
        # index of [sheet_name] sheet
        idx = writer.book.sheetnames.index(sheet_name)
        # remove [sheet_name]
        writer.book.remove(writer.book.worksheets[idx])
        # create an empty sheet [sheet_name] using old index
        writer.book.create_sheet(sheet_name, idx)

    # copy existing sheets
    writer.sheets = {ws.title: ws for ws in writer.book.worksheets}

    if startrow is None:
        startrow = 0

    # write out the new sheet
    df.to_excel(writer, sheet_name, startrow=startrow, **to_excel_kwargs)

    # save the workbook
    writer.save()


def save_to_excel(dict_df, file_name, dir_path):
    """
    保存结果至xlsx文件
    keys of dict 为 sheet name
    values of dict 为 worksheet table

    Parameters
    ----------
    dict_df : dict[str, pd.DataFrame]
    file_name : str
    dir_path : Path or str

    Returns
    -------
    """

    dir_path = Path(dir_path)
    dir_path.mkdir(parents=True, exist_ok=True)
    file_path = dir_path.joinpath(file_name)

    for key in dict_df:
        append_df_to_excel(file_path,
                           dict_df[key],
                           sheet_name=key,
                           index=False,
                           encoding='utf-8')



import pandas as pd
from sqlalchemy import select

from nuc_data_tool.db.base import Session
from nuc_data_tool.db.db_model import Nuc, NucData, File, PhysicalQuantity
from nuc_data_tool.db.db_utils import upsert
from nuc_data_tool.utils.middle_steps import middle_steps_line_serialization


def populate_database(xml_file):
    """
    将xml_file的数据填入数据库

    Parameters
    ----------
    xml_file: InputXmlFileReader

    Returns
    -------

    """
    session = Session()

    # 依据文件名获取对应的File object
    file_stmt = (select(File)
                 .where(File.name == xml_file.out_name)
                 )
    file_tmp = session.execute(file_stmt).scalar_one_or_none()
    if file_tmp is None:
        # 如果数据库不存在对应的File records则插入
        file_tmp = File(name=xml_file.out_name,
                        time_interval=xml_file.time_interval,
                        repeat_times=xml_file.repeat_times,
                        is_all_step=xml_file.is_all_step)
        session.add(file_tmp)
    else:
        session.close()
        return

    for key in xml_file.table_of_physical_quantity:

        if not xml_file.table_of_physical_quantity[key]:
            # 为空则跳过
            continue

        # 依据物理量名获取对应的PhysicalQuantity object
        physical_quantity_stmt = (select(PhysicalQuantity)
                                  .where(PhysicalQuantity.name == key)
                                  )
        physical_quantity_tmp = session.execute(physical_quantity_stmt).scalar_one_or_none()
        if physical_quantity_tmp is None:
            # 如果数据库不存在对应的PhysicalQuantity records则插入
            physical_quantity_tmp = PhysicalQuantity(name=key)
            session.add(physical_quantity_tmp)

        # 关系插入
        file_tmp.physical_quantities.append(physical_quantity_tmp)
        physical_quantity_tmp.files.append(file_tmp)

        # 将序列化的middle_steps_line和其他数据存入DataFrame
        # 如核素为gamma,则从0依次赋予nuc_ix
        df_all_tmp = pd.DataFrame(middle_steps_line_serialization(data.split())
                                  if key != 'gamma_spectra'
                                  else middle_steps_line_serialization([i, *data.split()])
                                  for i, data in enumerate(xml_file.table_of_physical_quantity[key])
                                  )

        # 截取核素部分(nuc_ix和name)
        df_nuc_tmp: pd.DataFrame = df_all_tmp.iloc[:, [0, 1]]
        df_nuc_tmp.columns = ('nuc_ix', 'name')

        # upsert into db
        stmt = upsert(Nuc,
                      df_nuc_tmp.to_dict(orient='records'),
                      update_field=df_nuc_tmp.columns.tolist(),
                      engine=session.bind)

        session.execute(stmt)

        # 如果len(df_all_tmp.columns)为5则说明有middle_steps列
        # 将数据部分截取出来
        if len(df_all_tmp.columns) != 5:
            df_data_tmp: pd.DataFrame = df_all_tmp.iloc[:, [-2, -1]]
            df_data_tmp.columns = ('first_step', 'last_step')
        else:
            df_data_tmp: pd.DataFrame = df_all_tmp.iloc[:, -3:]
            df_data_tmp.columns = ('first_step', 'last_step', 'middle_steps')

        list_nuc_id = session.execute(select(Nuc.id).
                                      where(Nuc.nuc_ix.in_(df_nuc_tmp['nuc_ix']))).scalars().all()

        # 为数据部分生成3个外键
        df_data_prefix = pd.DataFrame({'nuc_id': list_nuc_id,
                                       'file_id': file_tmp.id,
                                       'physical_quantity_id': physical_quantity_tmp.id})
        # 合并外键和数据部分
        df_data_all = pd.concat([df_data_prefix, df_data_tmp],
                                axis=1, copy=False)

        # almost twice as slow as __table__.insert
        # session.execute(insert(NucData).values(df_data_all.to_dict(orient='records')))
        session.execute(NucData.__table__.insert(), df_data_all.to_dict(orient='records'))

        session.commit()

    session.close()



from sqlalchemy import insert, delete
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.postgresql import insert as postgres_insert

from nuc_data_tool.db.base import Session, Base


def init_db():
    """
    初始化数据库

    Returns
    -------

    """
    with Session() as session:
        Base.metadata.drop_all(session.bind)
        Base.metadata.create_all(session.bind)


def delete_all_from_table(model):
    """
    删除某表的全部records

    Returns
    -------

    """
    with Session() as session:
        stmt = (delete(model))
        session.execute(stmt)
        session.commit()


def upsert(model, data, update_field, engine):
    """
    upsert实现
    依据session.bind.dialect得到当前数据库类型
    然后生成对应的upsert语句，当前支持mysql，postgresql，sqlite三种数据库
    on_duplicate_key_update for mysql
    on_conflict_do_nothing for postgresql
    insert or ignore for sqlite

    Parameters
    ----------
    model : Base
        orm model
    data : list
    update_field : list
    engine
        _engine.Engine instance
    Returns
    -------
    insert
        insert statement
    """

    if engine.dialect.name == 'mysql':
        stmt = mysql_insert(model).values(data)
        d = {f: getattr(stmt.inserted, f) for f in update_field}
        return stmt.on_duplicate_key_update(**d)
    elif engine.dialect.name == 'postgresql':
        stmt = postgres_insert(model).values(data)
        return stmt.on_conflict_do_nothing(index_elements=[update_field[0]])
    elif engine.dialect.name == 'sqlite':
        stmt = insert(model).values(data).prefix_with('OR IGNORE')
        return stmt
    else:
        raise Exception(f"can't support {engine.dialect.name} dialect")



"""
Database model

       ┌───────┐
       │       │
       │  nuc  │
       │       │
       └───────┘
           ▲ one
           │
           │
           │many
      ┌────┴─────┐           ┌─────────┐
      │          │many    one│         │
      │ nuc_data ├──────────►│  files  │
      │          │           │         │
      └────┬─────┘           └────┬────┘
           │many                  │many
           │                      │
           │                      │
           ▼ one                  │
┌───────────────────────┐         │
│                       │         │
│  physical_quantities  ├─────────┘
│                       │many
└───────────────────────┘

"""

from sqlalchemy import Column, Integer, Numeric, String, LargeBinary, Interval, Boolean, ForeignKey, Table
from sqlalchemy.orm import relationship

from nuc_data_tool.db.base import Base

files_physical_quantities_association = Table('files_physical_quantities_association', Base.metadata,
                                              Column('file_id', Integer, ForeignKey('files.id'), nullable=False),
                                              Column('physical_quantity_id', Integer,
                                                     ForeignKey('physical_quantities.id'), nullable=False)
                                              )


class Nuc(Base):
    __tablename__ = 'nuc'
    id = Column(Integer, primary_key=True)
    nuc_ix = Column(Integer, unique=True, autoincrement=True, nullable=False)
    name = Column(String(32), nullable=False)

    data = relationship('NucData', back_populates='nuc')


class NucData(Base):
    __tablename__ = 'nuc_data'
    id = Column(Integer, primary_key=True)
    nuc_id = Column(Integer, ForeignKey('nuc.id'), nullable=False)
    file_id = Column(Integer, ForeignKey('files.id'), nullable=False)
    physical_quantity_id = Column(Integer, ForeignKey('physical_quantities.id'), nullable=False)
    first_step = Column(Numeric(25), nullable=False)
    last_step = Column(Numeric(25), nullable=False)
    middle_steps = Column(LargeBinary)

    nuc = relationship('Nuc', back_populates='data')
    file = relationship('File', back_populates='data')
    physical_quantity = relationship('PhysicalQuantity', back_populates='data')


class File(Base):
    __tablename__ = 'files'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    time_interval = Column(Interval)
    repeat_times = Column(Integer)
    is_all_step = Column(Boolean)

    data = relationship('NucData', back_populates='file')
    physical_quantities = relationship('PhysicalQuantity',
                                       secondary=files_physical_quantities_association,
                                       back_populates='files')


class PhysicalQuantity(Base):
    __tablename__ = 'physical_quantities'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(16), nullable=False)

    data = relationship('NucData', back_populates='physical_quantity')
    files = relationship('File', secondary=files_physical_quantities_association,
                         back_populates='physical_quantities')



from pathlib import Path

import pandas as pd

from nuc_data_tool.db.db_model import File, PhysicalQuantity
from nuc_data_tool.db.fetch_data import (fetch_data_by_filename_and_nuclide_list, fetch_files_by_name,
                                         fetch_extracted_data_by_filename_and_physical_quantity,
                                         fetch_physical_quantities_by_name)
from nuc_data_tool.utils.formatter import type_checker
from nuc_data_tool.utils.workbook import save_to_excel


def filter_data(filename, physical_quantity_name, nuclide_list, is_all_step):
    """
    依据核素名称，筛选数据行
    :param is_all_step:
    :param filename: File object
    :param physical_quantity_name: str or list or PhysicalQuantity object
    :param nuclide_list: list, 关键核素，如果为None，则取不为0的全部核素
    :return: DataFrame，关键核素的计算结果
    """
    # if (nuclide_list is not None) and (len(nuclide_list) > 1):
    #     for key in dict_df_data:
    #         if key == 'gamma_spectra':
    #             continue
    #         is_in_nuclide_list = dict_df_data[key]['nuc_name'].isin(nuclide_list)
    #         dict_df_data[key]: pd.DataFrame = dict_df_data[key].loc[is_in_nuclide_list]
    # else:
    #     # Drop rows with all zeros in data.
    #     for key in dict_df_data:
    #         """
    #         The two lines below, does the exact same thing here.
    #         Drop rows with all zeros. But the latter one is way much faster.
    #         df_filter = df_data.iloc[df_data.apply(np.sum, axis=1).to_numpy().nonzero()]
    #         df_filter = df_data.iloc[df_data.any(axis=1).to_numpy().nonzero()]
    #
    #         When only condition is provided,
    #         the numpy.where() function is a shorthand for np.asarray(condition).nonzero().
    #         Using nonzero directly should be preferred, as it behaves correctly for subclasses.
    #         The rest of this documentation covers only the case where all three arguments are provided.
    #         But to be clear, I leave the np.where() kind of function in this comment.
    #         df_is_not_zero = np.where(df_density.any(axis=1))
    #         """
    #
    #         df_is_not_zero = dict_df_data[key][['first_step', 'last_step']].any(axis=1).to_numpy().nonzero()
    #         dict_df_data[key] = dict_df_data[key].iloc[df_is_not_zero]
    #         # df_output = df_data.loc[(df_density != Decimal(0)).any(axis=1), :]
    dict_df_data = fetch_data_by_filename_and_nuclide_list(filename, physical_quantity_name, nuclide_list, is_all_step)

    return dict_df_data


def save_extracted_data_to_exel(nuc_data_id,
                                filenames=None,
                                physical_quantities=None,
                                is_all_step=False,
                                result_path=Path('.'),
                                merge=True):
    """
    将数据存入到exel文件
    将传入的File list中包含的文件的数据存到exel文件
    如无filenames is None，则包含所有文件

    Parameters
    ----------
    nuc_data_id : list[int]
    filenames : list[File or str] or File or str
    physical_quantities : list[str or PhysicalQuantity] or str or PhysicalQuantity
        物理量，可以是物理量名的list[str]或str，
        也可以是list[PhysicalQuantity]或PhysicalQuantity
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列
    result_path : Path or str
    merge : bool, default = True
        是否将结果合并输出至一个文件，否则单独输出至每个文件

    Returns
    -------

    """

    if type_checker(filenames, File) == 'str':
        filenames = fetch_files_by_name(filenames)

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    file_name = 'final.xlsx'

    if is_all_step:
        file_name = f'all_steps_{file_name}'

    if merge:
        Path(result_path).joinpath(file_name).unlink(missing_ok=True)
    else:
        for filename in filenames:
            if is_all_step:
                Path(result_path).joinpath(f'all_steps_{filename.name}.xlsx').unlink(missing_ok=True)
            else:
                Path(result_path).joinpath(f'{filename.name}.xlsx').unlink(missing_ok=True)
        del filename

    physical_quantity: PhysicalQuantity
    for physical_quantity in physical_quantities:
        df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

        filename: File
        for filename in filenames:

            files_name = f'{filename.name}.xlsx'
            if is_all_step:
                files_name = f'all_steps_{filename.name}.xlsx'

            df_right = fetch_extracted_data_by_filename_and_physical_quantity(nuc_data_id,
                                                                              filename,
                                                                              physical_quantity,
                                                                              is_all_step)

            if not df_right.empty:
                df_left = pd.merge(df_left, df_right, how='outer', on=['nuc_ix', 'name'])

            if not merge:
                save_to_excel({physical_quantity.name: df_left},
                              files_name,
                              result_path)
                df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

        if merge:
            save_to_excel({physical_quantity.name: df_left},
                          file_name,
                          result_path)



from decimal import Decimal

from nuc_data_tool.utils.middle_steps_pb2 import MiddleStep, MiddleSteps


def serialization(middle_steps_list):
    middle_step = MiddleStep()
    middle_steps = MiddleSteps()
    for i, data in enumerate(middle_steps_list, start=1):
        middle_step.id = i
        middle_step.data = data
        middle_steps.middle_steps.append(middle_step)
    return middle_steps.SerializeToString()


def parsing(middle_steps_str):
    middle_steps = MiddleSteps()
    middle_steps.ParseFromString(middle_steps_str)
    return (middle_step for middle_step in middle_steps.middle_steps)


def middle_steps_line_serialization(data):
    """
    将middle_steps_line序列化
    len(data)小于10的话说明没有中间步骤，直接返回
    Parameters
    ----------
    data : list[str]

    Returns
    -------
    list[str]
    """
    if len(data) < 10:
        return data
    middle_steps = serialization(data[3: -1])
    return [*data[0:3], data[-1], middle_steps]


def middle_steps_line_parsing(data):
    """
    将middle_steps_line反序列化，并返回一个含序号和数据的字典

    Parameters
    ----------
    data : list[str]

    Returns
    -------
    dict
    """
    if data is None:
        return {'middle_steps': None}
    else:
        return {f'middle_step_{middle_step.id}': Decimal(middle_step.data) for middle_step in parsing(data)}



from pathlib import Path

import pandas as pd
from pycaret.anomaly import setup, create_model, predict_model, load_model

from nuc_data_tool.db.db_model import File, PhysicalQuantity
from nuc_data_tool.db.fetch_data import (fetch_files_by_name,
                                         fetch_physical_quantities_by_name,
                                         fetch_data_by_filename_and_physical_quantity)
from nuc_data_tool.utils.formatter import type_checker
from nuc_data_tool.utils.workbook import save_to_excel


def _complement_columns(nuc_data,
                        middle_steps_num):
    """
    补全 middle_step column 以便用于模型预测

    Parameters
    ----------
    nuc_data : pd.DataFrame
    middle_steps_num : int

    Returns
    -------
    pd.DataFrame
    """

    nuc_data_middle_steps_column_length = len(nuc_data.columns) - 4

    if middle_steps_num > nuc_data_middle_steps_column_length:
        complement_columns = [f'middle_step_{i}'
                              for i in range(nuc_data_middle_steps_column_length + 1,
                                             middle_steps_num + 1)]
        complement_df = pd.DataFrame(data=None, columns=complement_columns)
        nuc_data = pd.concat([nuc_data, complement_df], axis=1, copy=False)

    return nuc_data


def train_model(nuc_data,
                model_type,
                fraction):
    """

    Parameters
    ----------
    nuc_data : pd.DataFrame
    model_type : str
    fraction : float
    Returns
    -------

    """
    unnecessary_columns = ['nuc_ix', 'name']
    numeric_columns = [col for col in nuc_data.columns.tolist()
                       if col not in unnecessary_columns]

    exp_ano = setup(nuc_data,
                    normalize=True,
                    normalize_method='robust',
                    ignore_features=unnecessary_columns,
                    numeric_features=numeric_columns,
                    silent=True)

    kargs = {}
    if fraction is not None:
        kargs['fraction'] = fraction

    model = create_model(model_type, **kargs)

    return model


def prediction(filenames,
               physical_quantity='isotope',
               is_all_step=False,
               model_type='iforest',
               model=None,
               fraction=0.01):
    """

    Parameters
    ----------
    filenames : list[File or str] or File or str
    physical_quantity : str or PhysicalQuantity, default = 'isotope'
        物理量，可以是物理量名的list[str]或str，
        默认为核素密度
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列
    model_type : str
    model
    fraction : float

    Returns
    -------
    pd.DataFrame
    """

    if type_checker(filenames, File) == 'str':
        filenames = fetch_files_by_name(filenames)

    if type_checker(physical_quantity, PhysicalQuantity) == 'str':
        physical_quantity = fetch_physical_quantities_by_name(physical_quantity).pop()

    nuc_data_left = pd.DataFrame(columns=['nuc_ix', 'name'])

    for filename in filenames:
        nuc_data_right = fetch_data_by_filename_and_physical_quantity(filename, physical_quantity, is_all_step)

        if nuc_data_right.empty:
            continue

        nuc_data_right.rename(columns={'first_step': f'{filename.name}_first_step',
                                       'last_step': f'{filename.name}_last_step'},
                              inplace=True)
        columns = {col: f'{filename.name}_{col}'
                   for col in nuc_data_right.columns.tolist()
                   if 'middle_step' in col}
        nuc_data_right.rename(columns=columns, inplace=True)

        numeric_columns = [col for col in nuc_data_right.columns.tolist()
                           if col not in ['nuc_ix', 'name']]
        nuc_data_right[numeric_columns] = nuc_data_right[numeric_columns].astype('float64', copy=False)

        nuc_data_left = pd.merge(nuc_data_left, nuc_data_right, how='outer', on=['nuc_ix', 'name'])

    if model_type is not None:
        model = train_model(nuc_data=nuc_data_left, model_type=model_type, fraction=fraction)

    result_prediction = predict_model(model, data=nuc_data_left)

    return result_prediction[result_prediction['Anomaly'] == 1]


def save_prediction_to_exel(filenames,
                            result_path,
                            physical_quantities='isotope',
                            is_all_step=False,
                            merge=True,
                            model_type=None,
                            model_name=None,
                            fraction=0.001):
    """

    Parameters
    ----------
    filenames : list[File or str] or File or str
    physical_quantities : list[str or PhysicalQuantity] or str or PhysicalQuantity
        物理量，可以是物理量名的list[str]或str，
        也可以是list[PhysicalQuantity]或PhysicalQuantity
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列
    result_path : Path or str
    merge : bool, default = True
        是否将结果合并输出至一个文件，否则单独输出至每个文件
    model_type : str
    model_name : str
    fraction
    Returns
    -------

    """

    if type_checker(filenames, File) == 'str':
        filenames = fetch_files_by_name(filenames)

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    if model_type is None:
        model = load_model(model_name)
    else:
        model = None

    result_path = Path(result_path).joinpath('anomaly_detection_result')

    prefix = model_type

    file_name = 'final.xlsx'

    if is_all_step:
        file_name = f'all_steps_{file_name}'

    file_name = f'{prefix}_{file_name}'

    if merge:
        Path(result_path).joinpath(file_name).unlink(missing_ok=True)
    else:
        for filename in filenames:
            if is_all_step:
                Path(result_path).joinpath(f'{prefix}_all_steps_{filename.name}.xlsx').unlink(missing_ok=True)
            else:
                Path(result_path).joinpath(f'{prefix}_{filename.name}.xlsx').unlink(missing_ok=True)
        del filename

    for physical_quantity in physical_quantities:

        if merge:
            df_result = prediction(filenames=filenames,
                                   physical_quantity=physical_quantity,
                                   is_all_step=is_all_step,
                                   model_type=model_type,
                                   model=model,
                                   fraction=fraction)

            df_result.dropna(axis=1, how='all', inplace=True)
            save_to_excel({physical_quantity.name: df_result},
                          file_name,
                          result_path)
        else:
            df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])

            for filename in filenames:

                files_name = f'{prefix}_{filename.name}.xlsx'
                if is_all_step:
                    files_name = f'{prefix}_all_steps_{filename.name}.xlsx'

                df_right = prediction(filenames=filenames,
                                      physical_quantity=physical_quantity,
                                      is_all_step=is_all_step,
                                      model_type=model_type,
                                      model=model,
                                      fraction=fraction)

                if not df_right.empty:
                    df_right.drop(columns='Anomaly', inplace=True)
                    df_right.rename(columns={'Anomaly_Score': f'{filename.name}_Anomaly_Score'},
                                    inplace=True)

                    df_right.dropna(axis=1, how='all', inplace=True)
                    df_left = pd.merge(df_left, df_right, how='outer', on=['nuc_ix', 'name'])

                save_to_excel({physical_quantity.name: df_left},
                              files_name,
                              result_path)
                df_left = pd.DataFrame(data=None, columns=['nuc_ix', 'name'])


# def main():
#     """"""
#     from nuc_data_tool.db.fetch_data import fetch_files_by_name
#     files = fetch_files_by_name()
#
#     print(prediction(filenames=files,
#                      is_all_step=True,
#                      model_type='lof',
#                      fraction=0.01
#                      ))
#
#
# main()



from sqlalchemy import create_engine
from sqlalchemy.orm import registry
from sqlalchemy.orm import sessionmaker

from nuc_data_tool.utils.configlib import config


def _chosen_db(db_type=None, debug=False):
    """
    选择数据库（目前支持 Mysql, Postgresql, sqlite），生成对应的Session和engine

    Parameters
    ----------
    db_type : str
        数据库类型
    debug : bool, default False
        是否开启echo
    Returns
    -------
    tuple
        返回对应数据裤的Session和engine
    """
    db_config = config.get_database_config()

    if db_type is None:
        db_type = db_config['chosen_db']
    elif db_type not in db_config:
        db_type = 'sqlite'

    user = db_config[db_type].get('user')
    password = db_config[db_type].get('password')
    url = db_config[db_type].get('url')
    port = db_config[db_type].get('port')
    db_name = db_config[db_type].get('dbname')
    # sqlite
    path = db_config[db_type].get('path')

    if db_type == 'sqlite':
        connector_string = f'sqlite:///{path}'
        engine_tmp = create_engine(connector_string, future=True, echo=debug)
    elif db_type == 'mysql':
        connector_string = f'mysql+mysqlconnector://{user}:{password}@{url}:{port}/{db_name}?charset=utf8mb4'
        engine_tmp = create_engine(connector_string, future=True, echo=debug)
    elif db_type == 'postgresql':
        connector_string = f'postgresql+psycopg2://{user}:{password}@{url}:{port}/{db_name}?client_encoding=utf8'
        engine_tmp = create_engine(connector_string, future=True,
                                   executemany_mode='values',
                                   executemany_values_page_size=10000,
                                   executemany_batch_page_size=500,
                                   echo=debug)
    else:
        raise Exception(f"can't support {db_type} now")

    session_tmp = sessionmaker(bind=engine_tmp, future=True)

    return engine_tmp, session_tmp


mapper_registry = registry()
Base = mapper_registry.generate_base()
engine, Session = _chosen_db(debug=False)



from typing import Iterable

from nuc_data_tool.db.db_model import PhysicalQuantity, File
from nuc_data_tool.utils.configlib import config


def _get_physical_quantity_list_by_conf_file():
    """
    从配置文件获取全部物理量名
    Returns
    -------
    list
        全部物理量
    """
    return [key for key in config.get_data_extraction_conf('keys_of_rows')]


# 从配置文件获取全部物理量名
all_physical_quantity_list = _get_physical_quantity_list_by_conf_file()


def physical_quantity_list_generator(physical_quantities):
    """
    根据输入生成对应的physical_quantity list,如果输入为all，返回全部物理量list

    Parameters
    ----------
    physical_quantities : str or list[str]

    Returns
    -------
    list[str]
    """
    if isinstance(physical_quantities, str):
        # 如果输入的是str,则将其包装为list
        physical_quantities = [physical_quantities]

    if not all(physical_quantity in all_physical_quantity_list
               for physical_quantity in set(physical_quantities)) \
            and physical_quantities != 'all':

        # 遍历输入(physical_quantities)，如果不是每个element都在all_physical_quantity_list内
        # 并且输入不是all，设置输入为all

        physical_quantities = 'all'

    # 如果输入为all，返回全部物理量list，否，
    # 则先用set()去掉重复物理量名，然后再return 物理量list
    return all_physical_quantity_list \
        if physical_quantities == 'all' \
        else list(dict.fromkeys(physical_quantities).keys())


def type_checker(object_or_list, expected_type):
    """
    判断输入的是不是str or PhysicalQuantity or File or list[str or PhysicalQuantity or File]
    Parameters
    ----------
    object_or_list : str or PhysicalQuantity or File or list[str or PhysicalQuantity or File]
    expected_type : Any

    Returns
    -------
    str
    """

    if isinstance(object_or_list, Iterable):
        if all(isinstance(ele, str) for ele in object_or_list):
            return 'str'
        elif all(isinstance(ele, expected_type) for ele in object_or_list):
            return 'original'
        else:
            raise Exception('unexpected type')
    else:
        if isinstance(object_or_list, str):
            return 'str'
        elif isinstance(object_or_list, expected_type):
            return 'original'
        else:
            raise Exception('unexpected type')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: middle-steps.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='middle-steps.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12middle-steps.proto\"&\n\nMiddleStep\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t\"0\n\x0bMiddleSteps\x12!\n\x0cmiddle_steps\x18\x01 \x03(\x0b\x32\x0b.MiddleStepb\x06proto3'
)




_MIDDLESTEP = _descriptor.Descriptor(
  name='MiddleStep',
  full_name='MiddleStep',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='MiddleStep.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='MiddleStep.data', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=60,
)


_MIDDLESTEPS = _descriptor.Descriptor(
  name='MiddleSteps',
  full_name='MiddleSteps',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='middle_steps', full_name='MiddleSteps.middle_steps', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=110,
)

_MIDDLESTEPS.fields_by_name['middle_steps'].message_type = _MIDDLESTEP
DESCRIPTOR.message_types_by_name['MiddleStep'] = _MIDDLESTEP
DESCRIPTOR.message_types_by_name['MiddleSteps'] = _MIDDLESTEPS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MiddleStep = _reflection.GeneratedProtocolMessageType('MiddleStep', (_message.Message,), {
  'DESCRIPTOR' : _MIDDLESTEP,
  '__module__' : 'middle_steps_pb2'
  # @@protoc_insertion_point(class_scope:MiddleStep)
  })
_sym_db.RegisterMessage(MiddleStep)

MiddleSteps = _reflection.GeneratedProtocolMessageType('MiddleSteps', (_message.Message,), {
  'DESCRIPTOR' : _MIDDLESTEPS,
  '__module__' : 'middle_steps_pb2'
  # @@protoc_insertion_point(class_scope:MiddleSteps)
  })
_sym_db.RegisterMessage(MiddleSteps)


# @@protoc_insertion_point(module_scope)



from decimal import localcontext, Decimal, InvalidOperation
from pathlib import Path

import numpy as np
import pandas as pd

from nuc_data_tool.db.db_model import PhysicalQuantity, File
from nuc_data_tool.db.fetch_data import (fetch_extracted_data_by_filename_and_physical_quantity,
                                         fetch_files_by_name,
                                         fetch_physical_quantities_by_name)
from nuc_data_tool.utils.formatter import type_checker
from nuc_data_tool.utils.workbook import save_to_excel


def _complement_columns(df_reference,
                        df_comparison,
                        reference_complement_column_name,
                        comparison_complement_column_name):
    """
    对齐middle_step_* columns，数值填充为NaN

    Parameters
    ----------
    df_reference : pd.DataFrame
    df_comparison : pd.DataFrame
    reference_complement_column_name : str
    comparison_complement_column_name : str

    Returns
    -------
    tuple[pd.DataFrame, pd.DataFrame]
    """
    reference_column_length = len(df_reference.columns)
    comparison_column_length = len(df_comparison.columns)
    column_length_difference = reference_column_length - comparison_column_length

    column_start_num = min(reference_column_length, comparison_column_length) - 2

    if reference_column_length == comparison_column_length:
        # 当column length相同时，什么也不做
        pass
    elif reference_column_length < comparison_column_length:
        complement_columns = [f'{reference_complement_column_name}_middle_step_{i}'
                              for i in range(column_start_num,
                                             column_start_num + abs(column_length_difference))]
        complement_df = pd.DataFrame(data=None, columns=complement_columns)
        df_reference = pd.concat([df_reference, complement_df], axis=1, copy=False)
    else:
        complement_columns = [f'{comparison_complement_column_name}_middle_step_{i}'
                              for i in range(column_start_num,
                                             column_start_num + abs(column_length_difference))]
        complement_df = pd.DataFrame(data=None, columns=complement_columns)
        df_comparison = pd.concat([df_comparison, complement_df], axis=1, copy=False)

    return df_reference, df_comparison


def _calculate_deviation(df_reference,
                         df_comparison,
                         deviation_mode='relative',
                         threshold=Decimal('1.0E-12')):
    """
    计算误差
    relative deviation formula: abs(X - Y) / 1 + min(abs(X), abs(Y))
    absolute deviation formula: X - Y

    Parameters
    ----------
    df_reference : pd.DataFrame
    df_comparison : pd.DataFrame
    deviation_mode : str, default = 'relative'
        绝对=absolute
        相对=relative
        偏差模式，分为绝对和相对，默认为相对

    Returns
    -------
    tuple[pd.DataFrame, pd.Series]
    """

    df_deviation = pd.DataFrame()
    for reference_column, \
        comparison_column in zip(df_reference.columns.values.tolist()[2:],
                                 df_comparison.columns.values.tolist()[2:]):
        i = 0
        if deviation_mode == 'relative':
            # abs(X - Y) / 1 + min(abs(X), abs(Y))
            """NaN is classified as unordered.
            By default InvalidOperation is trapped, 
            so a Python exception is raised when using <= and >= against Decimal('NaN'). 
            This is a logical extension; 
            Python has actual exceptions so if you compare against the 
            NaN exception value, you can expect an exception being raised. 
            You could disable trapping by using a Decimal.localcontext()
            https://stackoverflow.com/a/28371465/11071374
            """
            with localcontext() as ctx:
                ctx.traps[InvalidOperation] = False
                deviation = (df_reference[reference_column] - df_comparison[comparison_column]).abs() / \
                            (1 + np.minimum(df_reference[reference_column].fillna(Decimal('NaN')),
                                            df_comparison[comparison_column].fillna(Decimal('NaN'))))

        elif deviation_mode == 'absolute':
            deviation = (df_reference[reference_column] - df_comparison[comparison_column]).abs()

        else:
            raise Exception("wrong deviation mode")

        if not deviation.isnull().values.all():
            df_deviation[f'relative_deviation_middle_step_{i}'] = deviation

        i += 1

    df_deviation.rename(columns={'relative_deviation_middle_step_0': 'relative_deviation_last_step'}, inplace=True)

    reserved_index = (df_deviation.T > threshold).any()
    df_deviation = df_deviation.loc[reserved_index].reset_index(drop=True)

    return df_deviation, reserved_index


def _merge_reference_comparison_and_deviation(df_reference,
                                              df_comparison,
                                              df_deviation,
                                              reserved_index,
                                              reserved_na=False):
    """
    依据reserved_index合并reference，comparison和deviation表
    如果reserved_na为False,则drop columns with all NaN's，否则反之
    默认reserved_na为False

    Parameters
    ----------
    df_reference : pd.DataFrame
    df_comparison : pd.DataFrame
    df_deviation : pd.DataFrame
    reserved_index : pd.Series
    reserved_na : bool, default = False
        如果reserved_na为False,则drop columns with all NaN's，否则反之

    Returns
    -------
    pd.DataFrame
    """
    df_ = pd.merge(df_reference.loc[reserved_index],
                   df_comparison.loc[reserved_index],
                   how='outer', on=['nuc_ix', 'name'])

    df_all = pd.concat([df_, df_deviation], axis=1, copy=False)

    if not reserved_na:
        df_all.dropna(axis=1, how='all', inplace=True)

    return df_all


def calculate_comparative_result(nuc_data_id,
                                 reference_file,
                                 comparison_file,
                                 physical_quantities='isotope',
                                 deviation_mode='relative',
                                 threshold=Decimal('1.0E-12'),
                                 is_all_step=False):
    """
    选定一个基准文件，一个对比文件，与其进行对比，计算并返回对比结果

    Parameters
    ----------
    nuc_data_id : list[int]
    reference_file : File or str
        基准文件
    comparison_file : File or str
        对比文件
    physical_quantities : list[str or PhysicalQuantity] or str or PhysicalQuantity, default = 'isotope'
        对比用物理量，可以是物理量名的list[str]或str，
        也可以是PhysicalQuantity list也可以是list[PhysicalQuantity]或PhysicalQuantity
        默认为核素密度
    deviation_mode : str, default = 'relative'
        绝对=absolute
        相对=relative
        偏差模式，分为绝对和相对，默认为相对
    threshold : Decimal, default = Decimal('1.0E-12')
        偏差阈值，默认1.0E-12
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------
    dict[str, pd.DataFrame]
    """

    if type_checker([reference_file, comparison_file], File) == 'str':
        reference_file = fetch_files_by_name(reference_file).pop()
        comparison_file = fetch_files_by_name(comparison_file).pop()

    if type_checker(physical_quantities, PhysicalQuantity) == 'str':
        physical_quantities = fetch_physical_quantities_by_name(physical_quantities)

    dict_df_all = {}

    physical_quantity: PhysicalQuantity
    for physical_quantity in physical_quantities:
        reference_data = fetch_extracted_data_by_filename_and_physical_quantity(nuc_data_id,
                                                                                reference_file,
                                                                                physical_quantity,
                                                                                is_all_step)

        comparison_data = fetch_extracted_data_by_filename_and_physical_quantity(nuc_data_id,
                                                                                 comparison_file,
                                                                                 physical_quantity,
                                                                                 is_all_step
                                                                                 )

        if reference_data.empty or comparison_data.empty:
            continue

        reference_data, comparison_data = _complement_columns(reference_data,
                                                              comparison_data,
                                                              reference_file.name,
                                                              comparison_file.name)

        df_deviation, reserved_index = _calculate_deviation(reference_data,
                                                            comparison_data,
                                                            deviation_mode,
                                                            Decimal(threshold))

        dict_df_all[physical_quantity.name] = _merge_reference_comparison_and_deviation(reference_data,
                                                                                        comparison_data,
                                                                                        df_deviation,
                                                                                        reserved_index)

    return dict_df_all


def save_comparison_result_to_excel(nuc_data_id,
                                    reference_file,
                                    comparison_files,
                                    result_path,
                                    physical_quantities='isotope',
                                    deviation_mode='relative',
                                    threshold=Decimal('1.0E-12'),
                                    is_all_step=False):
    """
    选定一个基准文件，使其与对比文件列表中的文件一一对比，计算并输出对比结果至工作簿(xlsx文件)

    Parameters
    ----------
    nuc_data_id : list[int]
    reference_file : File or str
        基准文件
    comparison_files : list[str or File]or File or str
        对比文件列表
    result_path : Path or str
    physical_quantities : list[str or PhysicalQuantity] or str or PhysicalQuantity, default = 'isotope'
        对比用物理量，可以是物理量名的list[str]或str，
        也可以是PhysicalQuantity list也可以是list[PhysicalQuantity]或PhysicalQuantity
        默认为核素密度
    deviation_mode : str, default = 'relative'
        绝对=absolute
        相对=relative
        偏差模式，分为绝对和相对，默认为相对
    threshold : Decimal, default = Decimal('1.0E-12')
        偏差阈值，默认1.0E-12
    is_all_step : bool, default = False
        是否读取全部中间结果数据列，默认只读取最终结果列

    Returns
    -------
    """

    if type_checker(reference_file, File) == 'str':
        reference_file = fetch_files_by_name(reference_file)

    if type_checker(comparison_files, File) == 'str':
        comparison_files = fetch_files_by_name(comparison_files)

    for comparison_file in comparison_files:
        print((reference_file.name, comparison_file.name))

        dict_df_all = calculate_comparative_result(nuc_data_id=nuc_data_id,
                                                   reference_file=reference_file,
                                                   comparison_file=comparison_file,
                                                   physical_quantities=physical_quantities,
                                                   deviation_mode=deviation_mode,
                                                   threshold=threshold,
                                                   is_all_step=is_all_step)

        file_name = f'{deviation_mode}_{threshold}_{reference_file.name}_vs_{comparison_file.name}.xlsx'

        if is_all_step:
            file_name = f'all_step_{file_name}'

        Path(result_path).joinpath('comparative_result').joinpath(file_name).unlink(missing_ok=True)
        save_to_excel(dict_df_all,
                      file_name,
                      Path(result_path).joinpath('comparative_result'))



import linecache
from datetime import timedelta
from pathlib import Path
from xml.etree.ElementTree import parse

from nuc_data_tool.utils.configlib import config
from nuc_data_tool.utils.formatter import physical_quantity_list_generator


class InputXmlFileReader:
    """
    表示xml.out文件的类

    Attributes
    ----------
    out_path : Path
        xml.out文件路径
    physical_quantities : str
            核素名

    """
    chosen_physical_quantity: list
    length_of_physical_quantity: dict

    def __init__(self, out_path, physical_quantities='all'):
        """
        可以根据输入的文件路径和物理量
        自动计算得出选择的物理量chosen_physical_quantity和
        物理量对应的行号范围length_of_physical_quantity

        Parameters
        ----------
        out_path : Path or str
            xml.out文件路径
        physical_quantities : str or list[str]
            核素名
        """
        self.time_interval = None
        self.repeat_times = None
        self.is_all_step = None
        self.parent_path = Path(out_path).parent
        self.xml_name = Path(out_path).stem
        self.xml_path = self.parent_path.joinpath(self.xml_name)
        self.out_name = self.xml_path.stem
        self.out_path = Path(out_path)
        self.chosen_physical_quantity = physical_quantity_list_generator(physical_quantities)
        self.length_of_physical_quantity = self.get_length_of_physical_quantity(self.chosen_physical_quantity)
        self.unfetched_physical_quantity = self.get_unfetched_physical_quantity()
        self.fetched_physical_quantity = self.get_fetched_physical_quantity()
        self.table_of_physical_quantity = self.get_table_of_physical_quantity()
        self.set_file_info(self.xml_path)

    def __enter__(self):
        """
        with statement
        Returns
        -------
        InputXmlFileReader
        """
        self.file_object = self.out_path.open(mode='r', encoding='UTF-8')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file_object.close()

    def __getitem__(self, physical_quantity_name):
        tmp_dict = self.table_of_physical_quantity
        if physical_quantity_name != 'all':
            return {physical_quantity_name: tmp_dict.get(physical_quantity_name)}
        else:
            return self.table_of_physical_quantity

    def set_chosen_physical_quantity(self, physical_quantity_name):
        """
        输入选择的物理量，设置类的属性chosen_physical_quantity和length_of_physical_quantity
        Parameters
        ----------
        physical_quantity_name : str
            物理量名
        Returns
        -------

        """
        self.chosen_physical_quantity = physical_quantity_list_generator(physical_quantity_name)
        self.length_of_physical_quantity = self.get_length_of_physical_quantity()

    def set_file_info(self, xml_path):
        try:
            root = parse(xml_path)
        except FileNotFoundError as e:
            print(f"{self.xml_name} doesn't find")
            print(e)
            print()
            return

        burn = root.find('burnup/burn').attrib

        if burn['unit'] == 'year':
            self.time_interval = timedelta(days=365 * float(burn['time']))
        elif burn['unit'] == 'day':
            self.time_interval = timedelta(days=float(burn['time']))
        else:
            pass

        self.repeat_times = burn['repeat']

        self.is_all_step = bool(root.find('output/table').attrib['print_all_step'])

    def get_length_of_physical_quantity(self, physical_quantities=None):
        """
        获取对应物理量的行号范围，physical_quantity_name默认为None,设置则输出physical_quantity_name对应的行号的字典
        但是此只是临时返回一个物理量及行号范围的字典，不修改类的属性self.chosen_physical_quantity和self.length_of_physical_quantity
        Parameters
        ----------
        physical_quantities: str or list[str]
            物理量名

        Returns
        -------
        dict[str, list[int]]
        """

        physical_quantities = physical_quantity_list_generator(physical_quantities)

        # 开始搜索关键字
        index_start = {physical_quantity: config.get_data_extraction_conf("keys_of_rows").get(physical_quantity)[0]
                       for physical_quantity in physical_quantities}
        # 结尾搜索关键字
        index_end = {physical_quantity: config.get_data_extraction_conf("keys_of_rows").get(physical_quantity)[-1]
                     for physical_quantity in physical_quantities}

        length_of_physical_quantity = {key: [] for key in physical_quantities}

        for physical_quantity in physical_quantities:
            is_find_start_title = True
            for row_number, line in enumerate(self.out_path.open(encoding='UTF-8')):
                if is_find_start_title:
                    if index_start[physical_quantity] in line:
                        length_of_physical_quantity[physical_quantity].append(
                            row_number + 7
                            if physical_quantity != 'gamma_spectra'
                            else
                            row_number + 2)
                        is_find_start_title = False
                else:
                    if index_end[physical_quantity] in line:
                        length_of_physical_quantity[physical_quantity].append(
                            row_number - 3
                            if physical_quantity != 'gamma_spectra'
                            else
                            row_number - 2)
                        break

        return length_of_physical_quantity

    def get_unfetched_physical_quantity(self):
        """
        获取在选取范围内却未能成功获取的物理量

        Returns
        -------
        list
        """
        unfetched_physical_quantity = [name for name in self.chosen_physical_quantity
                                       if not self.length_of_physical_quantity.get(name)]
        return unfetched_physical_quantity

    def get_fetched_physical_quantity(self):
        """
        获取在选取范围内成功获取的物理量

        Returns
        -------
        list
        """
        fetched_physical_quantity = [name for name in self.chosen_physical_quantity
                                     if self.length_of_physical_quantity.get(name)]
        return fetched_physical_quantity

    def get_table_of_physical_quantity(self):
        """
        依据选择的物理量和未能成功获取的物理量，以及对应物理量的行号范围
        从xml.out文件中使用linecache.getlines获取文本内容
        Returns
        -------
        dict[str, list[str]]
        """
        df_data = {}
        for key in self.chosen_physical_quantity:
            text = []
            if key not in self.unfetched_physical_quantity:
                row_start = self.length_of_physical_quantity[key][0]
                row_end = self.length_of_physical_quantity[key][1]
                text = linecache.getlines(str(self.out_path))[row_start:row_end + 1]

            df_data[key] = text
        return df_data



from pathlib import Path

import toml


class Config:
    """
    配置文件类
    """

    def __init__(self, path):
        try:
            config_file_path = Path(path)
            self.conf = toml.load(config_file_path)
        except FileNotFoundError:
            print("The config file doesn't exist!")

    def get_conf(self, properties):
        """
        获取整个配置文件

        Parameters
        ----------
        properties : str
            字段

        Returns
        -------
        dict
        """
        return self.conf.get(properties)

    def get_data_extraction_conf(self, properties):
        """
        获取data_extraction下的配置

        Parameters
        ----------
        properties : str
            字段

        Returns
        -------
        dict
        """
        return self.conf.get("data_extraction").get(properties)

    def get_file_path(self, properties):
        """
        获取文件路径

        Parameters
        ----------
        properties : str
            字段

        Returns
        -------
        Path
        """
        return Path(self.conf.get("file_path").get(properties))

    def get_anomaly_detection_config(self, properties):
        """
        获取异常检测的配置
        Returns
        -------

        """
        if isinstance(self.conf.get("anomaly_detection").get(properties), str):
            return Path(self.conf.get("anomaly_detection").get(properties))
        else:
            return self.conf.get("anomaly_detection").get(properties)

    def get_database_config(self):
        """
        获取database下的配置

        Returns
        -------
        dict
        """
        return self.conf.get("database")

    def get_nuclide_list(self, nuclide_name):
        """
        输入核素列表名，获取对应的核素列表
        Parameters
        ----------
        nuclide_name : str
            核素列表名
        Returns
        -------
        list[str]
        """
        return self.conf.get("nuclide_list").get(nuclide_name, None)


CONFIG_FILE_PATH = Path.cwd().joinpath('config.toml')
if not CONFIG_FILE_PATH.exists():
    CONFIG_FILE_PATH = Path(__file__).parent.parent.joinpath('config.toml')

config = Config(CONFIG_FILE_PATH)



import pathlib
import re

from setuptools import find_packages
from setuptools import setup

from nuc_data_tool import __version__


def read(filename):
    filename = pathlib.Path.joinpath(pathlib.Path(__file__).parent, filename)
    text_type = type(u"")
    with filename.open(mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


setup(
    name="nuc-data-tool",
    version=__version__,
    url="https://github.com/bookyue/nuclear_data_automated_processing",
    license='LGPL',

    author="Kyle Yue",
    author_email="sevenbookyue@gmail.com",

    description="an automated processing tool for specific nuclear data",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",

    packages=find_packages(exclude=('tests',)),

    include_package_data=True,

    install_requires=["SQLAlchemy >= 1.4.0", "pandas", "toml",
                      "protobuf", "openpyxl", "click",
                      "psycopg2", "mysql-connector-python", "pycaret >= 2.3.0"],
    python_requires=">=3.8",

    entry_points={
        "console_scripts": [
            "nuctool=nuc_data_tool.__main__:main",
        ]
    },

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
    ],
)
